@@ -1,6 +1,4 @@
-# -*- coding: utf-8 -*-
-#
-# Copyright 2010-2017 The pygit2 contributors
+# Copyright 2010-2020 The pygit2 contributors
 #
 # This file is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License, version 2,
@@ -25,45 +23,42 @@
 # the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 # Boston, MA 02110-1301, USA.
 
-# Import from the future
-from __future__ import absolute_import
-
 # Import from the Standard Library
+from io import BytesIO
 from string import hexdigits
-import sys, tarfile
+import tarfile
 from time import time
-
-if sys.version_info[0] < 3:
-    from cStringIO import StringIO
-else:
-    from io import BytesIO as StringIO
-
-import six
+import warnings
 
 # Import from pygit2
-from _pygit2 import Repository as _Repository, init_file_backend
-from _pygit2 import Oid, GIT_OID_HEXSZ, GIT_OID_MINPREFIXLEN
-from _pygit2 import GIT_CHECKOUT_SAFE, GIT_CHECKOUT_RECREATE_MISSING, GIT_DIFF_NORMAL
-from _pygit2 import GIT_FILEMODE_LINK
-from _pygit2 import GIT_BRANCH_LOCAL, GIT_BRANCH_REMOTE, GIT_BRANCH_ALL
-from _pygit2 import Reference, Tree, Commit, Blob, Signature
+from ._pygit2 import Repository as _Repository, init_file_backend
+from ._pygit2 import Oid, GIT_OID_HEXSZ, GIT_OID_MINPREFIXLEN
+from ._pygit2 import GIT_CHECKOUT_SAFE, GIT_CHECKOUT_RECREATE_MISSING, GIT_DIFF_NORMAL
+from ._pygit2 import GIT_FILEMODE_LINK
+from ._pygit2 import GIT_BRANCH_LOCAL, GIT_BRANCH_REMOTE, GIT_BRANCH_ALL
+from ._pygit2 import GIT_REF_SYMBOLIC
+from ._pygit2 import Reference, Tree, Commit, Blob
+from ._pygit2 import InvalidSpecError
 
+from .callbacks import git_fetch_options
 from .config import Config
 from .errors import check_error
 from .ffi import ffi, C
 from .index import Index
 from .remote import RemoteCollection
 from .blame import Blame
-from .utils import to_bytes, is_string
+from .utils import to_bytes, StrArray
 from .submodule import Submodule
 
 
 class BaseRepository(_Repository):
-    def __init__(self, backend, *args, **kwargs):
-        super(BaseRepository, self).__init__(backend, *args, **kwargs)
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
         self._common_init()
 
     def _common_init(self):
+        self.branches = Branches(self)
+        self.references = References(self)
         self.remotes = RemoteCollection(self)
 
         # Get the pointer as the contents of a buffer and store it for
@@ -72,6 +67,25 @@ class BaseRepository(_Repository):
         ffi.buffer(repo_cptr)[:] = self._pointer[:]
         self._repo = repo_cptr[0]
 
+    # Backwards compatible ODB access
+    def read(self, *args, **kwargs):
+        """read(oid) -> type, data, size
+
+        Read raw object data from the repository.
+        """
+        return self.odb.read(*args, **kwargs)
+
+    def write(self, *args, **kwargs):
+        """write(type, data) -> Oid
+
+        Write raw object data into the repository. First arg is the object
+        type, the second one a buffer with data. Return the Oid of the created
+        object."""
+        return self.odb.write(*args, **kwargs)
+
+    def __iter__(self):
+        return iter(self.odb)
+
     def lookup_submodule(self, path):
         csub = ffi.new('git_submodule **')
         cpath = ffi.new('char[]', to_bytes(path))
@@ -80,6 +94,23 @@ class BaseRepository(_Repository):
         check_error(err)
         return Submodule._from_c(self, csub[0])
 
+    def update_submodules(self, submodules=None, init=False, callbacks=None):
+        if submodules is None:
+            submodules = self.listall_submodules()
+
+        # prepare options
+        opts = ffi.new('git_submodule_update_options *')
+        C.git_submodule_update_init_options(opts, C.GIT_SUBMODULE_UPDATE_OPTIONS_VERSION)
+
+        with git_fetch_options(callbacks, opts=opts.fetch_opts) as payload:
+            i = 1 if init else 0
+            for submodule in submodules:
+                submodule_instance = self.lookup_submodule(submodule)
+                err = C.git_submodule_update(submodule_instance._subm, i, opts)
+                payload.check_error(err)
+
+        return None
+
     #
     # Mapping interface
     #
@@ -103,10 +134,7 @@ class BaseRepository(_Repository):
     # Remotes
     #
     def create_remote(self, name, url):
-        """Create a new remote. Return a <Remote> object.
-
-        This method is deprecated, please use Remote.remotes.create()
-        """
+        warnings.warn("Use repo.remotes.create(..)", DeprecationWarning)
         return self.remotes.create(name, url)
 
     #
@@ -172,11 +200,35 @@ class BaseRepository(_Repository):
 
         return self.create_reference_symbolic(name, target, force)
 
+    def resolve_refish(self, refish):
+        """Convert a reference-like short name "ref-ish" to a valid
+        (commit, reference) pair.
+
+        If ref-ish points to a commit, the reference element of the result
+        will be None.
+
+        Examples::
+
+            repo.resolve_refish('mybranch')
+            repo.resolve_refish('sometag')
+            repo.resolve_refish('origin/master')
+            repo.resolve_refish('bbb78a9')
+        """
+        try:
+            reference = self.lookup_reference_dwim(refish)
+        except (KeyError, InvalidSpecError):
+            reference = None
+            commit = self.revparse_single(refish)
+        else:
+            commit = reference.peel(Commit)
+
+        return (commit, reference)
+
     #
     # Checkout
     #
     @staticmethod
-    def _checkout_args_to_options(strategy=None, directory=None):
+    def _checkout_args_to_options(strategy=None, directory=None, paths=None):
         # Create the options struct to pass
         copts = ffi.new('git_checkout_options *')
         check_error(C.git_checkout_init_options(copts, 1))
@@ -195,6 +247,11 @@ class BaseRepository(_Repository):
             refs.append(target_dir)
             copts.target_directory = target_dir
 
+        if paths:
+            strarray = StrArray(paths)
+            refs.append(strarray)
+            copts.paths = strarray.array[0]
+
         return copts, refs
 
     def checkout_head(self, **kwargs):
@@ -205,13 +262,13 @@ class BaseRepository(_Repository):
         copts, refs = Repository._checkout_args_to_options(**kwargs)
         check_error(C.git_checkout_head(self._repo, copts))
 
-    def checkout_index(self, **kwargs):
-        """Checkout the repository's index
+    def checkout_index(self, index=None, **kwargs):
+        """Checkout the given index or the repository's index
 
         For arguments, see Repository.checkout().
         """
         copts, refs = Repository._checkout_args_to_options(**kwargs)
-        check_error(C.git_checkout_index(self._repo, ffi.NULL, copts))
+        check_error(C.git_checkout_index(self._repo, index._index if index else ffi.NULL, copts))
 
     def checkout_tree(self, treeish, **kwargs):
         """Checkout the given treeish
@@ -226,29 +283,36 @@ class BaseRepository(_Repository):
 
     def checkout(self, refname=None, **kwargs):
         """
-        Checkout the given reference using the given strategy, and update
-        the HEAD.
+        Checkout the given reference using the given strategy, and update the
+        HEAD.
         The reference may be a reference name or a Reference object.
         The default strategy is GIT_CHECKOUT_SAFE | GIT_CHECKOUT_RECREATE_MISSING.
 
-        To checkout from the HEAD, just pass 'HEAD'::
-
-          >>> checkout('HEAD')
-
-        This is identical to calling checkout_head().
-
         If no reference is given, checkout from the index.
 
-        Arguments:
+        Parameters:
 
-        :param str|Reference refname: The reference to checkout. After checkout,
-          the current branch will be switched to this one.
+        refname : str or Reference
+            The reference to checkout. After checkout, the current branch will
+            be switched to this one.
 
-        :param int strategy: A ``GIT_CHECKOUT_`` value. The default is
-          ``GIT_CHECKOUT_SAFE``.
+        strategy : int
+            A ``GIT_CHECKOUT_`` value. The default is ``GIT_CHECKOUT_SAFE``.
 
-        :param str directory: Alternative checkout path to workdir.
+        directory : str
+            Alternative checkout path to workdir.
 
+        paths : list[str]
+            A list of files to checkout from the given reference.
+            If paths is provided, HEAD will not be set to the reference.
+
+        Examples:
+
+        * To checkout from the HEAD, just pass 'HEAD'::
+
+            >>> checkout('HEAD')
+
+          This is identical to calling checkout_head().
         """
 
         # Case 1: Checkout index
@@ -270,18 +334,20 @@ class BaseRepository(_Repository):
         treeish = self[oid]
         self.checkout_tree(treeish, **kwargs)
 
-        self.set_head(refname)
+        if 'paths' not in kwargs:
+            self.set_head(refname)
 
     #
     # Setting HEAD
     #
     def set_head(self, target):
-        """Set HEAD to point to the given target
+        """
+        Set HEAD to point to the given target.
 
-        Arguments:
+        Parameters:
 
         target
-            The new target for HEAD. Can be a string or Oid (to detach)
+            The new target for HEAD. Can be a string or Oid (to detach).
         """
 
         if isinstance(target, Oid):
@@ -298,6 +364,30 @@ class BaseRepository(_Repository):
     #
     # Diff
     #
+    def __whatever_to_tree_or_blob(self, obj):
+        if obj is None:
+            return None
+
+        # If it's a string, then it has to be valid revspec
+        if isinstance(obj, str):
+            obj = self.revparse_single(obj)
+        elif isinstance(obj, Oid):
+            obj = self[obj]
+
+        # First we try to get to a blob
+        try:
+            obj = obj.peel(Blob)
+        except Exception:
+            # And if that failed, try to get a tree, raising a type
+            # error if that still doesn't work
+            try:
+                obj = obj.peel(Tree)
+            except Exception:
+                raise TypeError('unexpected "%s"' % type(obj))
+
+        return obj
+
+
     def diff(self, a=None, b=None, cached=False, flags=GIT_DIFF_NORMAL,
              context_lines=3, interhunk_lines=0):
         """
@@ -322,19 +412,21 @@ class BaseRepository(_Repository):
             Otherwise the referred object is compared to 'a'
 
         cached
-            if 'b' is None, by default the working directory is compared to 'a'.
+            If 'b' is None, by default the working directory is compared to 'a'.
             If 'cached' is set to True, the index/staging area is used for comparing.
 
         flag
-            a GIT_DIFF_* constant
+            A combination of GIT_DIFF_* constants. For a list of the constants,
+            with a description, see git_diff_option_t in
+            https://github.com/libgit2/libgit2/blob/master/include/git2/diff.h
 
         context_lines
-            the number of unchanged lines that define the boundary
-            of a hunk (and to display before and after)
+            The number of unchanged lines that define the boundary of a hunk
+            (and to display before and after)
 
         interhunk_lines
-            the maximum number of unchanged lines between hunk
-            boundaries before the hunks will be merged into a one
+            The maximum number of unchanged lines between hunk boundaries
+            before the hunks will be merged into a one
 
         Examples::
 
@@ -357,29 +449,8 @@ class BaseRepository(_Repository):
         API (Tree.diff_to_tree()) directly.
         """
 
-        def whatever_to_tree_or_blob(obj):
-            if obj is None:
-                return None
-
-            # If it's a string, then it has to be valid revspec
-            if is_string(obj):
-                obj = self.revparse_single(obj)
-
-            # First we try to get to a blob
-            try:
-                obj = obj.peel(Blob)
-            except Exception:
-                # And if that failed, try to get a tree, raising a type
-                # error if that still doesn't work
-                try:
-                    obj = obj.peel(Tree)
-                except Exception:
-                    raise TypeError('unexpected "%s"' % type(obj))
-
-            return obj
-
-        a = whatever_to_tree_or_blob(a)
-        b = whatever_to_tree_or_blob(b)
+        a = self.__whatever_to_tree_or_blob(a)
+        b = self.__whatever_to_tree_or_blob(b)
 
         opt_keys = ['flags', 'context_lines', 'interhunk_lines']
         opt_values = [flags, context_lines, interhunk_lines]
@@ -418,29 +489,36 @@ class BaseRepository(_Repository):
     def blame(self, path, flags=None, min_match_characters=None,
               newest_commit=None, oldest_commit=None, min_line=None,
               max_line=None):
-        """Return a Blame object for a single file.
+        """
+        Return a Blame object for a single file.
 
-        Arguments:
+        Parameters:
 
         path
             Path to the file to blame.
+
         flags
             A GIT_BLAME_* constant.
+
         min_match_characters
             The number of alphanum chars that must be detected as moving/copying
             within a file for it to associate those lines with the parent commit.
+
         newest_commit
             The id of the newest commit to consider.
+
         oldest_commit
-          The id of the oldest commit to consider.
+            The id of the oldest commit to consider.
+
         min_line
             The first line in the file to blame.
+
         max_line
             The last line in the file to blame.
 
         Examples::
 
-            repo.blame('foo.c', flags=GIT_BLAME_TRACK_COPIES_SAME_FILE)");
+            repo.blame('foo.c', flags=GIT_BLAME_TRACK_COPIES_SAME_FILE)
         """
 
         options = ffi.new('git_blame_options *')
@@ -474,7 +552,7 @@ class BaseRepository(_Repository):
         """Index representing the repository's index file."""
         cindex = ffi.new('git_index **')
         err = C.git_repository_index(cindex, self._repo)
-        check_error(err, True)
+        check_error(err, io=True)
 
         return Index.from_c(self, cindex)
 
@@ -544,14 +622,19 @@ class BaseRepository(_Repository):
         return ret
 
     def merge_commits(self, ours, theirs, favor='normal'):
-        """Merge two arbitrary commits
+        """
+        Merge two arbitrary commits.
 
-        Arguments:
+        Returns: an index with the result of the merge.
+
+        Parameters:
 
         ours
             The commit to take as "ours" or base.
+
         theirs
             The commit which will be merged into "ours"
+
         favor
             How to deal with file-level conflicts. Can be one of
 
@@ -560,22 +643,19 @@ class BaseRepository(_Repository):
             * theirs. The "theirs" side of the conflict region is used.
             * union. Unique lines from each side will be used.
 
-            for all but NORMAL, the index will not record a conflict.
+            For all but NORMAL, the index will not record a conflict.
 
         Both "ours" and "theirs" can be any object which peels to a commit or the id
         (string or Oid) of an object which peels to a commit.
-
-        Returns an index with the result of the merge
-
         """
 
         ours_ptr = ffi.new('git_commit **')
         theirs_ptr = ffi.new('git_commit **')
         cindex = ffi.new('git_index **')
 
-        if is_string(ours) or isinstance(ours, Oid):
+        if isinstance(ours, (str, Oid)):
             ours = self[ours]
-        if is_string(theirs) or isinstance(theirs, Oid):
+        if isinstance(theirs, (str, Oid)):
             theirs = self[theirs]
 
         ours = ours.peel(Commit)
@@ -592,27 +672,31 @@ class BaseRepository(_Repository):
         return Index.from_c(self, cindex)
 
     def merge_trees(self, ancestor, ours, theirs, favor='normal'):
-        """Merge two trees
+        """
+        Merge two trees.
 
-        Arguments:
+        Returns: an Index that reflects the result of the merge.
+
+        Parameters:
 
         ancestor
-            The tree which is the common ancestor between 'ours' and 'theirs'
+            The tree which is the common ancestor between 'ours' and 'theirs'.
+
         ours
             The commit to take as "ours" or base.
+
         theirs
-            The commit which will be merged into "ours"
+            The commit which will be merged into "ours".
+
         favor
-            How to deal with file-level conflicts. Can be one of
+            How to deal with file-level conflicts. Can be one of:
 
             * normal (default). Conflicts will be preserved.
             * ours. The "ours" side of the conflict region is used.
             * theirs. The "theirs" side of the conflict region is used.
             * union. Unique lines from each side will be used.
 
-            for all but NORMAL, the index will not record a conflict.
-
-        Returns an Index that reflects the result of the merge.
+            For all but NORMAL, the index will not record a conflict.
         """
 
         ancestor_ptr = ffi.new('git_tree **')
@@ -620,11 +704,11 @@ class BaseRepository(_Repository):
         theirs_ptr = ffi.new('git_tree **')
         cindex = ffi.new('git_index **')
 
-        if is_string(ancestor) or isinstance(ancestor, Oid):
+        if isinstance(ancestor, (str, Oid)):
             ancestor = self[ancestor]
-        if is_string(ours) or isinstance(ours, Oid):
+        if isinstance(ours, (str, Oid)):
             ours = self[ours]
-        if is_string(theirs) or isinstance(theirs, Oid):
+        if isinstance(theirs, (str, Oid)):
             theirs = self[theirs]
 
         ancestor = ancestor.peel(Tree)
@@ -650,34 +734,54 @@ class BaseRepository(_Repository):
                  only_follow_first_parent=None,
                  show_commit_oid_as_fallback=None, abbreviated_size=None,
                  always_use_long_format=None, dirty_suffix=None):
-        """Describe a commit-ish or the current working tree.
+        """
+        Describe a commit-ish or the current working tree.
 
-        :param committish: Commit-ish object or object name to describe, or
-            `None` to describe the current working tree.
-        :type committish: `str`, :class:`~.Reference`, or :class:`~.Commit`
+        Returns: The description (str).
 
-        :param int max_candidates_tags: The number of candidate tags to
-            consider. Increasing above 10 will take slightly longer but may
-            produce a more accurate result. A value of 0 will cause only exact
-            matches to be output.
-        :param int describe_strategy: A GIT_DESCRIBE_* constant.
-        :param str pattern: Only consider tags matching the given `glob(7)`
-            pattern, excluding the "refs/tags/" prefix.
-        :param bool only_follow_first_parent: Follow only the first parent
-            commit upon seeing a merge commit.
-        :param bool show_commit_oid_as_fallback: Show uniquely abbreviated
-            commit object as fallback.
-        :param int abbreviated_size: The minimum number of hexadecimal digits
-            to show for abbreviated object names. A value of 0 will suppress
-            long format, only showing the closest tag.
-        :param bool always_use_long_format: Always output the long format (the
-            nearest tag, the number of commits, and the abbrevated commit name)
-            even when the committish matches a tag.
-        :param str dirty_suffix: A string to append if the working tree is
-            dirty.
+        Parameters:
 
-        :returns: The description.
-        :rtype: `str`
+        committish : `str`, :class:`~.Reference`, or :class:`~.Commit`
+            Commit-ish object or object name to describe, or `None` to describe
+            the current working tree.
+
+        max_candidates_tags : int
+            The number of candidate tags to consider. Increasing above 10 will
+            take slightly longer but may produce a more accurate result. A
+            value of 0 will cause only exact matches to be output.
+
+        describe_strategy : int
+            Can be one of:
+
+            * `GIT_DESCRIBE_DEFAULT` - Only match annotated tags. (This is
+              equivalent to setting this parameter to `None`.)
+            * `GIT_DESCRIBE_TAGS` - Match everything under refs/tags/
+              (includes lightweight tags).
+            * `GIT_DESCRIBE_ALL` - Match everything under refs/ (includes
+              branches).
+
+        pattern : str
+            Only consider tags matching the given `glob(7)` pattern, excluding
+            the "refs/tags/" prefix.
+
+        only_follow_first_parent : bool
+            Follow only the first parent commit upon seeing a merge commit.
+
+        show_commit_oid_as_fallback : bool
+            Show uniquely abbreviated commit object as fallback.
+
+        abbreviated_size : int
+            The minimum number of hexadecimal digits to show for abbreviated
+            object names. A value of 0 will suppress long format, only showing
+            the closest tag.
+
+        always_use_long_format : bool
+            Always output the long format (the nearest tag, the number of
+            commits, and the abbrevated commit name) even when the committish
+            matches a tag.
+
+        dirty_suffix : str
+            A string to append if the working tree is dirty.
 
         Example::
 
@@ -704,7 +808,7 @@ class BaseRepository(_Repository):
 
         result = ffi.new('git_describe_result **')
         if committish:
-            if is_string(committish):
+            if isinstance(committish, str):
                 committish = self.revparse_single(committish)
 
             commit = committish.peel(Commit)
@@ -738,7 +842,7 @@ class BaseRepository(_Repository):
             try:
                 return ffi.string(buf.ptr).decode('utf-8')
             finally:
-                C.git_buf_free(buf)
+                C.git_buf_dispose(buf)
         finally:
             C.git_describe_result_free(result[0])
 
@@ -747,26 +851,36 @@ class BaseRepository(_Repository):
     #
     def stash(self, stasher, message=None, keep_index=False,
               include_untracked=False, include_ignored=False):
-        """Save changes to the working directory to the stash.
+        """
+        Save changes to the working directory to the stash.
 
-        :param Signature stasher: The identity of the person doing the stashing.
-        :param str message: An optional description of stashed state.
-        :param bool keep_index: Leave changes already added to the index
-            in the working directory.
-        :param bool include_untracked: Also stash untracked files.
-        :param bool include_ignored: Also stash ignored files.
+        Returns: The Oid of the stash merge commit (Oid).
 
-        :returns: The Oid of the stash merge commit.
-        :rtype: Oid
+        Parameters:
+
+        stasher : Signature
+            The identity of the person doing the stashing.
+
+        message : str
+            An optional description of stashed state.
+
+        keep_index : bool
+            Leave changes already added to the index in the working directory.
+
+        include_untracked : bool
+            Also stash untracked files.
+
+        include_ignored : bool
+            Also stash ignored files.
 
         Example::
 
-            >>> repo = pygit2.Repsitory('.')
+            >>> repo = pygit2.Repository('.')
             >>> repo.stash(repo.default_signature(), 'WIP: stashing')
         """
 
-        if message is not None:
-            stash_msg = ffi.new('char[]', to_bytes(message)) if message else ffi.NULL
+        if message:
+            stash_msg = ffi.new('char[]', to_bytes(message))
         else:
             stash_msg = ffi.NULL
 
@@ -798,18 +912,24 @@ class BaseRepository(_Repository):
         return stash_opts
 
     def stash_apply(self, index=0, **kwargs):
-        """Apply a stashed state in the stash list to the working directory.
+        """
+        Apply a stashed state in the stash list to the working directory.
 
-        :param int index: The position within the stash list of the stash to apply.
-            0 is the most recent stash.
-        :param bool reinstate_index: Try to reinstate stashed changes to the index.
+        Parameters:
+
+        index : int
+            The position within the stash list of the stash to apply. 0 is the
+            most recent stash.
+
+        reinstate_index : bool
+            Try to reinstate stashed changes to the index.
 
         The checkout options may be customized using the same arguments taken by
         Repository.checkout().
 
         Example::
 
-            >>> repo = pygit2.Repsitory('.')
+            >>> repo = pygit2.Repository('.')
             >>> repo.stash(repo.default_signature(), 'WIP: stashing')
             >>> repo.stash_apply(strategy=GIT_CHECKOUT_ALLOW_CONFLICTS)
         """
@@ -817,10 +937,14 @@ class BaseRepository(_Repository):
         check_error(C.git_stash_apply(self._repo, index, stash_opts))
 
     def stash_drop(self, index=0):
-        """Remove a stashed state from the stash list.
+        """
+        Remove a stashed state from the stash list.
 
-        :param int index: The position within the stash list of the stash to remove.
-            0 is the most recent stash.
+        Parameters:
+
+        index : int
+            The position within the stash list of the stash to remove. 0 is
+            the most recent stash.
         """
         check_error(C.git_stash_drop(self._repo, index))
 
@@ -836,7 +960,8 @@ class BaseRepository(_Repository):
     # Utility for writing a tree into an archive
     #
     def write_archive(self, treeish, archive, timestamp=None, prefix=''):
-        """Write treeish into an archive
+        """
+        Write treeish into an archive.
 
         If no timestamp is provided and 'treeish' is a commit, its committer
         timestamp will be used. Otherwise the current time will be used.
@@ -844,14 +969,17 @@ class BaseRepository(_Repository):
         All path names in the archive are added to 'prefix', which defaults to
         an empty string.
 
-        Arguments:
+        Parameters:
 
         treeish
             The treeish to write.
+
         archive
-            An archive from the 'tarfile' module
+            An archive from the 'tarfile' module.
+
         timestamp
             Timestamp to use for the files in the archive.
+
         prefix
             Extra prefix to add to the path names in the archive.
 
@@ -859,7 +987,7 @@ class BaseRepository(_Repository):
 
             >>> import tarfile, pygit2
             >>>> with tarfile.open('foo.tar', 'w') as archive:
-            >>>>     repo = pygit2.Repsitory('.')
+            >>>>     repo = pygit2.Repository('.')
             >>>>     repo.write_archive(repo.head.target, archive)
         """
 
@@ -867,7 +995,7 @@ class BaseRepository(_Repository):
         if isinstance(treeish, Tree):
             tree = treeish
 
-        if isinstance(treeish, Oid) or is_string(treeish):
+        if isinstance(treeish, (str, Oid)):
             treeish = self[treeish]
 
         # if we don't have a timestamp, try to get it from a commit
@@ -900,28 +1028,30 @@ class BaseRepository(_Repository):
                 info.size = 0
                 archive.addfile(info)
             else:
-                archive.addfile(info, StringIO(content))
+                info.mode = tree[entry.path].filemode
+                archive.addfile(info, BytesIO(content))
 
     #
     # Ahead-behind, which mostly lives on its own namespace
     #
     def ahead_behind(self, local, upstream):
-        """Calculate how many different commits are in the non-common parts
-        of the history between the two given ids.
+        """
+        Calculate how many different commits are in the non-common parts of the
+        history between the two given ids.
 
-        Ahead is how many commits are in the ancestry of the 'local'
-        commit which are not in the 'upstream' commit. Behind is the
-        opposite.
+        Ahead is how many commits are in the ancestry of the 'local' commit
+        which are not in the 'upstream' commit. Behind is the opposite.
 
-        Arguments
+        Returns: a tuple of two integers with the number of commits ahead and
+        behind respectively.
+
+        Parameters:
 
         local
-            The commit which is considered the local or current state
-        upstream
-            The commit which is considered the upstream
+            The commit which is considered the local or current state.
 
-        Returns a tuple of two integers with the number of commits ahead and
-        behind respectively.
+        upstream
+            The commit which is considered the upstream.
         """
 
         if not isinstance(local, Oid):
@@ -943,21 +1073,24 @@ class BaseRepository(_Repository):
     # Git attributes
     #
     def get_attr(self, path, name, flags=0):
-        """Retrieve an attribute for a file by path
+        """
+        Retrieve an attribute for a file by path.
 
-        Arguments
+        Returns: a boolean, None if the value is unspecified, or string with
+        the value of the attribute.
+
+        Parameters:
 
         path
             The path of the file to look up attributes for, relative to the
-            workdir root
+            workdir root.
+
         name
-            The name of the attribute to look up
+            The name of the attribute to look up.
+
         flags
             A combination of GIT_ATTR_CHECK_ flags which determine the
-            lookup order
-
-        Returns either a boolean, None (if the value is unspecified) or string
-        with the value of the attribute.
+            lookup order.
         """
 
         cvalue = ffi.new('char **')
@@ -1002,19 +1135,22 @@ class BaseRepository(_Repository):
         check_error(err)
 
     def revert_commit(self, revert_commit, our_commit, mainline=0):
-        """Reverts the given Commit against the given "our" Commit,
-           producing an Index that reflects the result of the revert.
+        """
+        Reverts the given Commit against the given "our" Commit, producing an
+        Index that reflects the result of the revert.
 
-        Arguments
+        Returns: an Index with the result of the revert.
+
+        Parameters:
 
         revert_commit
-            The Commit to revert
-        our_commit
-            The Commit to revert against (eg, HEAD)
-        mainline
-            The parent of the revert Commit, if it is a merge (i.e. 1, 2)
+            The Commit to revert.
 
-        Returns an Index with the result of the revert.
+        our_commit
+            The Commit to revert against (eg, HEAD).
+
+        mainline
+            The parent of the revert Commit, if it is a merge (i.e. 1, 2).
         """
         cindex = ffi.new('git_index **')
         revert_commit_ptr = ffi.new('git_commit **')
@@ -1035,14 +1171,21 @@ class BaseRepository(_Repository):
         return Index.from_c(self, cindex)
 
 
-class Branches(object):
-    def __init__(self, repository, flag=GIT_BRANCH_ALL):
+class Branches:
+
+    def __init__(self, repository, flag=GIT_BRANCH_ALL, commit=None):
         self._repository = repository
         self._flag = flag
+        if commit is not None:
+            if isinstance(commit, Commit):
+                commit = commit.id
+            elif not isinstance(commit, Oid):
+                commit = self._repository.expand_id(commit)
+        self._commit = commit
 
         if flag == GIT_BRANCH_ALL:
-            self.local = Branches(repository, flag=GIT_BRANCH_LOCAL)
-            self.remote = Branches(repository, flag=GIT_BRANCH_REMOTE)
+            self.local = Branches(repository, flag=GIT_BRANCH_LOCAL, commit=commit)
+            self.remote = Branches(repository, flag=GIT_BRANCH_REMOTE, commit=commit)
 
     def __getitem__(self, name):
         branch = None
@@ -1052,7 +1195,7 @@ class Branches(object):
         if branch is None and self._flag & GIT_BRANCH_REMOTE:
             branch = self._repository.lookup_branch(name, GIT_BRANCH_REMOTE)
 
-        if branch is None:
+        if branch is None or not self._valid(branch):
             raise KeyError('Branch not found: {}'.format(name))
 
         return branch
@@ -1065,7 +1208,8 @@ class Branches(object):
 
     def __iter__(self):
         for branch_name in self._repository.listall_branches(self._flag):
-            yield branch_name
+            if self._commit is None or self.get(branch_name) is not None:
+                yield branch_name
 
     def create(self, name, commit, force=False):
         return self._repository.create_branch(name, commit, force)
@@ -1073,11 +1217,26 @@ class Branches(object):
     def delete(self, name):
         self[name].delete()
 
+    def _valid(self, branch):
+        if branch.type == GIT_REF_SYMBOLIC:
+            branch = branch.resolve()
+
+        return (
+            self._commit is None
+            or branch.target == self._commit
+            or self._repository.descendant_of(branch.target, self._commit)
+        )
+
+    def with_commit(self, commit):
+        assert self._commit is None
+        return Branches(self._repository, self._flag, commit)
+
     def __contains__(self, name):
         return self.get(name) is not None
 
 
-class References(object):
+class References:
+
     def __init__(self, repository):
         self._repository = repository
 
@@ -1107,17 +1266,23 @@ class References(object):
     def objects(self):
         return self._repository.listall_reference_objects()
 
+    def compress(self):
+        return self._repository.compress_references()
+
 
 class Repository(BaseRepository):
-    def __init__(self, path, *args, **kwargs):
-        if not isinstance(path, six.string_types):
-            path = path.decode('utf-8')
-
-        path_backend = init_file_backend(path)
-        super(Repository, self).__init__(backend=path_backend, *args, **kwargs)
-
-        self.branches = Branches(self)
-        self.references = References(self)
+    def __init__(self, *args, **kwargs):
+        if len(args) != 0:
+            path = args[0]
+            args = args[1:]
+            if hasattr(path, "__fspath__"):
+                path = path.__fspath__()
+            if not isinstance(path, str):
+                path = path.decode('utf-8')
+            path_backend = init_file_backend(path)
+            super().__init__(path_backend, *args, **kwargs)
+        else:
+            super().__init__(*args, **kwargs)
 
     @classmethod
     def _from_c(cls, ptr, owned):
